#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <sys/timeb.h>

#include "FPtree.h"
#include "inline_routine.h"

FPtree goFPtree;

int *gplocal_item_sup_map;
int *gplocal_item_sumsup_map;
int** gpptailitem_buf;
int gntail_buf_pages;
int gntail_buf_size;
int gnpage_no;
int gnpage_pos;

int gncounttail_calls;
int gncountsup_calls;
double gdtail_count_time;


void FPtree::InsertTransaction(FP_NODE* &proot, HEADER_TABLE pheader_table, int* ptransaction, int length, int nlast_item_pos, int frequency, TID_NODE *ptid_head, TID_NODE *ptid_tail)
{
	FP_NODE *pcur_node, *pnew_node, *pparent, *pleftsib;
	int i, j;

	pcur_node = proot;
	pparent = NULL;
	pleftsib = NULL;
	for(i=0;i<length;i++)
	{
		while(pcur_node!=NULL && pcur_node->nitem_order<ptransaction[i])
		{
			pleftsib = pcur_node;
			pcur_node = pcur_node->prightsibling;
		}

		if(pcur_node==NULL || pcur_node->nitem_order>ptransaction[i])
		{
			pnew_node = NewOneFPNode();
			pnew_node->nitem_order = ptransaction[i];
			pnew_node->frequency = frequency;
			if(i<length-1)
				pnew_node->nchild_freq = frequency;
			else
				pnew_node->nchild_freq = 0;
			pnew_node->pchild = NULL;
			pnew_node->pparent = pparent;
			pnew_node->prightsibling = pcur_node;
			if(pleftsib!=NULL)
				pleftsib->prightsibling = pnew_node;
			else if(pparent!=NULL)
				pparent->pchild = pnew_node;
			pnew_node->pnode_link = pheader_table[pnew_node->nitem_order].pconddb;
			pheader_table[pnew_node->nitem_order].pconddb = pnew_node;
			if(i==0 && pleftsib==NULL)
				proot = pnew_node;
			pparent = pnew_node;
			if(i==nlast_item_pos)
			{
				pnew_node->ptid_head = ptid_head;
				pnew_node->ptid_tail = ptid_tail;
			}

			for(j=i+1;j<length;j++)
			{
				pnew_node = NewOneFPNode();
				pnew_node->nitem_order = ptransaction[j];
				pnew_node->frequency = frequency;
				if(j<length-1)
					pnew_node->nchild_freq = frequency;
				else
					pnew_node->nchild_freq = 0;
				pnew_node->pchild = NULL;
				pnew_node->pparent = pparent;
				pparent->pchild = pnew_node;
				pnew_node->prightsibling = NULL;
				pnew_node->pnode_link = pheader_table[pnew_node->nitem_order].pconddb;
				pheader_table[pnew_node->nitem_order].pconddb = pnew_node;
				pparent = pnew_node;
				if(j==nlast_item_pos)
				{
					pnew_node->ptid_head = ptid_head;
					pnew_node->ptid_tail = ptid_tail;
				}
			}
			break;
		}
		else 
		{
			pcur_node->frequency += frequency;
			if(i<length-1)
				pcur_node->nchild_freq += frequency;
			if(i==nlast_item_pos)
			{
				if(pcur_node->ptid_head==NULL)
					pcur_node->ptid_head = ptid_head;
				else
					pcur_node->ptid_tail->pnext = ptid_head;
				pcur_node->ptid_tail = ptid_tail;
			}
			pparent = pcur_node;
			pcur_node = pcur_node->pchild;
			pleftsib = NULL;
		}
	}
}

void FPtree::CountFreqItems(HEADER_TABLE pheader_table, int nitem_order, int *pitem_sup_map)
{
	FP_NODE *pitemnode, *pfpnode;
	int ncount;

	pitemnode = pheader_table[nitem_order].pconddb;
	while(pitemnode!=NULL)
	{
		ncount = pitemnode->frequency;
		pfpnode = pitemnode->pparent;
		while(pfpnode!=NULL)
		{
			pitem_sup_map[pfpnode->nitem_order] += ncount;
			pfpnode = pfpnode->pparent;
		}
		pitemnode = pitemnode->pnode_link;
	}
}


void FPtree::countsup(FP_NODE *pfpnode)
{
	FP_NODE *pchild_node;

	pchild_node = pfpnode->pchild;
	while(pchild_node!=NULL)
	{
		gplocal_item_sup_map[pchild_node->nitem_order] += pchild_node->frequency;
		if(pchild_node->nchild_freq==pchild_node->frequency)
			countsup(pchild_node);
		pchild_node = pchild_node->prightsibling;
	}
}

void FPtree::countsup(FP_NODE *pfpnode, HEADER_TABLE phead_table, int num_of_freqitems)
{
	FP_NODE *pchild_node;
	int i;

	pchild_node = pfpnode->pchild;
	while(pchild_node!=NULL)
	{
		gplocal_item_sup_map[pchild_node->nitem_order] += pchild_node->frequency;
		if(pchild_node->nchild_freq==pchild_node->frequency)
		{
			if(pchild_node->nitem_order>=num_of_freqitems)
				countsup(pchild_node);
			else if(!phead_table[pchild_node->nitem_order].bgen_fullitems)
				countsup(pchild_node, phead_table, num_of_freqitems);
			else if(pchild_node->ptail_itemset!=NULL)
			{
				for(i=1;i<=pchild_node->ptail_itemset[0];i++)
					gplocal_item_sup_map[pchild_node->ptail_itemset[i]] += pchild_node->frequency;
			}
		}
		pchild_node = pchild_node->prightsibling;
	}

}

int FPtree::CountTailFullItems(HEADER_TABLE phead_table, int ncur_order, int nfull_sup, int num_of_freqitems, int num_of_local_items)
{
	FP_NODE *pfpnode;
	int i, num_of_local_full, num_of_comm_items;
	struct timeb start, end;

	ftime(&start);
	gncounttail_calls++;

	phead_table[ncur_order].bgen_fullitems = true;
	
	memset(gplocal_item_sumsup_map, 0, sizeof(int)*num_of_local_items);

	pfpnode = phead_table[ncur_order].pconddb;
	while(pfpnode!=NULL)
	{
		if(pfpnode->nchild_freq==pfpnode->frequency)
		{
			memset(gplocal_item_sup_map, 0, sizeof(int)*num_of_local_items);
			gncountsup_calls++;
			countsup(pfpnode, phead_table, num_of_freqitems);

			num_of_local_full = 0;
			for(i=ncur_order+1;i<num_of_local_items;i++)
			{
				if(gplocal_item_sup_map[i]==pfpnode->frequency)
				{
					gplocal_item_sumsup_map[i] += gplocal_item_sup_map[i];
					num_of_local_full++;
				}
			}
			if(num_of_local_full>0)
			{
				pfpnode->ptail_itemset = InsertTailItemset(num_of_local_full);
				pfpnode->ptail_itemset[0] = num_of_local_full;
				num_of_local_full = 1;
				for(i=ncur_order+1;i<num_of_local_items;i++)
				{
					if(gplocal_item_sup_map[i]==pfpnode->frequency)
						pfpnode->ptail_itemset[num_of_local_full++] = i;
				}
			}
		}
		pfpnode = pfpnode->pnode_link;
	}

	num_of_comm_items = 0;
	for(i=ncur_order+1;i<num_of_local_items;i++)
	{
		if(gplocal_item_sumsup_map[i]==nfull_sup)
			num_of_comm_items++;
	}

	ftime(&end);
	gdtail_count_time += end.time-start.time+(double)(end.millitm-start.millitm)/1000;

	return num_of_comm_items;
}

int FPtree::GetTailFullItems(FP_NODE *pconddb, int nfull_sup, int nitem_order, int num_of_local_items)
{
	FP_NODE *pfpnode, *pitemnode;
	int num_of_comm_items, i, ntrans_len, ncur_order;
	bool bcontained;

	ncur_order = pconddb->nitem_order;
	memset(gplocal_item_sumsup_map, 0, sizeof(int)*num_of_local_items);
	num_of_comm_items = 0;

	pitemnode = pconddb;
	while(pitemnode!=NULL)
	{
		bcontained = false;
		ntrans_len = 0;
		pfpnode = pitemnode->pparent;
		while(pfpnode!=NULL)
		{
			if(pfpnode->nitem_order==nitem_order)
			{
				bcontained = true;
				break;
			}
			pfpnode = pfpnode->pparent;			
		}
		if(bcontained)
		{
			if(pitemnode->ptail_itemset!=NULL)
			{
				for(i=1;i<=pitemnode->ptail_itemset[0];i++)
					gplocal_item_sumsup_map[pitemnode->ptail_itemset[i]] += pitemnode->frequency;
			}
			else 
				break;
		}
		pitemnode = pitemnode->pnode_link;
	}
	for(i=ncur_order+1;i<num_of_local_items;i++)
	{
		if(gplocal_item_sumsup_map[i]==nfull_sup)
			num_of_comm_items++;
	}
	return num_of_comm_items;
}

FP_NODE* FPtree::BuildNewTreeWTTail(FP_NODE *pconddb, HEADER_TABLE pnewheader_table, int *pitem_order_map)
{
	FP_NODE *pnewroot, *pitemnode, *pfpnode;
	int ncount, ntrans_len, i, nlast_item_pos;

	pnewroot = NULL;

	pitemnode = pconddb;
	while(pitemnode!=NULL)
	{
		ncount = pitemnode->frequency;
		pfpnode = pitemnode->pparent;
		ntrans_len = 0;
		while(pfpnode!=NULL)
		{
			if(pitem_order_map[pfpnode->nitem_order]>=0)
				gptransaction[ntrans_len++] = pitem_order_map[pfpnode->nitem_order];
			pfpnode = pfpnode->pparent;
		}

		if(ntrans_len>0 && ntrans_len+gnprefix_len>=gnmin_size)
		{
			nlast_item_pos = ntrans_len-1;
			if(ntrans_len>1)
				sort_trans(gptransaction, ntrans_len);
			if(pitemnode->ptail_itemset!=NULL)
			{
				for(i=1;i<=pitemnode->ptail_itemset[0];i++)
				{
					if(pitem_order_map[pitemnode->ptail_itemset[i]]>=0)
						gptransaction[ntrans_len++] = pitem_order_map[pitemnode->ptail_itemset[i]];
				}
			}
			if(ntrans_len>1)
				InsertTransaction(pnewroot, pnewheader_table, gptransaction, ntrans_len, nlast_item_pos, ncount, pitemnode->ptid_head, pitemnode->ptid_tail);
			else if(ntrans_len==1)
			{
				if(pnewheader_table[gptransaction[0]].ptid_head==NULL)
					pnewheader_table[gptransaction[0]].ptid_head = pitemnode->ptid_head;
				else
					pnewheader_table[gptransaction[0]].ptid_tail->pnext = pitemnode->ptid_head;
				pnewheader_table[gptransaction[0]].ptid_tail = pitemnode->ptid_tail;
			}
		}
		pitemnode = pitemnode->pnode_link;
	}
	return pnewroot;
}


void FPtree::PushRightTidList(FP_NODE *pconddb)
{
	FP_NODE *pfpnode, *pparent;

	pfpnode = pconddb;
	while(pfpnode!=NULL)
	{
		pparent = pfpnode->pparent;
		if(pparent!=NULL)
		{
			if(pparent->ptid_tail==NULL)
				pparent->ptid_head = pfpnode->ptid_head;
			else
				pparent->ptid_tail->pnext = pfpnode->ptid_head;
			pparent->ptid_tail = pfpnode->ptid_tail;
		}
		pfpnode = pfpnode->pnode_link;
	}

}

int FPtree::GetTidList(int support, FP_NODE *pcond_db, int *ptid_list)
{
	if(gnprefix_len<gnmin_size || support<gnmin_sup)
		return 0;

	FP_NODE *pfpnode;
	TID_NODE *ptid_node;
	int nlist_len;

	nlist_len = 0;
	pfpnode = pcond_db;

	if(gntype==0 || gnprefix_len<support)
	{
		if(!gbresult_name_given)
			return support;

		while(pfpnode!=NULL)
		{
			ptid_node = pfpnode->ptid_head;
			while(ptid_node!=pfpnode->ptid_tail)
			{
				ptid_list[nlist_len++] = ptid_node->ntid;
				ptid_node = ptid_node->pnext;
			}
			ptid_list[nlist_len++] = ptid_node->ntid;

			pfpnode = pfpnode->pnode_link;
		}
	}
	else if(gnprefix_len==support)
	{
		while(pfpnode!=NULL)
		{
			ptid_node = pfpnode->ptid_head;
			while(1)
			{
				if(gpitem_order_map[ptid_node->ntid]>gnfirstitem_order)
					return 0;			
				ptid_list[nlist_len++] = ptid_node->ntid;
				if(ptid_node==pfpnode->ptid_tail)
					break;
				ptid_node = ptid_node->pnext;
			}
			pfpnode = pfpnode->pnode_link;
		}
	}
	else 
		printf("Error with length and support\n");

	if(nlist_len!=support)
		printf("Error: inconsistent support\n");

	return nlist_len;
}

int FPtree::GetTidList(int nitem, int support, FP_NODE *pcond_db, int* ptid_list)
{
	if(gnprefix_len<gnmin_size || support<gnmin_sup)
		return 0;

	FP_NODE *pitemnode, *pfpnode;
	TID_NODE *ptid_node;
	int nlist_len;

	nlist_len = 0;
	pitemnode = pcond_db;
	if(gntype==0 || gnprefix_len<support)
	{
		if(!gbresult_name_given)
			return support;

		while(pitemnode!=NULL)
		{
			pfpnode = pitemnode->pparent;
			while(pfpnode!=NULL)
			{
				if(pfpnode->nitem_order==nitem)
					break;
				pfpnode = pfpnode->pparent;
			}
			if(pfpnode!=NULL)
			{
				ptid_node = pitemnode->ptid_head;
				while(ptid_node!=pitemnode->ptid_tail)
				{
					ptid_list[nlist_len++] = ptid_node->ntid;
					ptid_node = ptid_node->pnext;
				}
				ptid_list[nlist_len++] = ptid_node->ntid;
			}
			pitemnode = pitemnode->pnode_link;
		}
	}
	else if(gnprefix_len==support)
	{
		while(pitemnode!=NULL)
		{
			pfpnode = pitemnode->pparent;
			while(pfpnode!=NULL)
			{
				if(pfpnode->nitem_order==nitem)
					break;
				pfpnode = pfpnode->pparent;
			}
			if(pfpnode!=NULL)
			{
				ptid_node = pitemnode->ptid_head;
				while(1)
				{
					if(gpitem_order_map[ptid_node->ntid]>gnfirstitem_order)
						return 0;
					ptid_list[nlist_len++] = ptid_node->ntid;
					if(ptid_node==pitemnode->ptid_tail)
						break;
					ptid_node = ptid_node->pnext;
				}
			}
			pitemnode = pitemnode->pnode_link;
		}
	}
	else
		printf("Error with length and support\n");


	if(nlist_len!=support)
		printf("Error: inconsistent support\n");

	return nlist_len;

}

int FPtree::GetTidList(int support, HEADER_NODE *pheader_node, int* ptid_list)
{
	if(gnprefix_len<gnmin_size || support<gnmin_sup)
		return 0;

	FP_NODE *pfpnode;
	TID_NODE *ptid_node;
	int nlist_len;

	nlist_len = 0;
	ptid_node = pheader_node->ptid_head;
	
	if(gntype==0 || gnprefix_len<support)
	{
		if(!gbresult_name_given)
			return support;

		while(ptid_node!=pheader_node->ptid_tail)
		{
			ptid_list[nlist_len++] = ptid_node->ntid;
			ptid_node = ptid_node->pnext;
		}
		if(ptid_node!=NULL)
			ptid_list[nlist_len++] = ptid_node->ntid;
		
		pfpnode = pheader_node->pconddb;
		while(pfpnode!=NULL )
		{
			ptid_node = pfpnode->ptid_head;
			while(ptid_node!=pfpnode->ptid_tail)
			{
				ptid_list[nlist_len++] = ptid_node->ntid;
				ptid_node = ptid_node->pnext;
			}
			ptid_list[nlist_len++] = ptid_node->ntid;
			pfpnode = pfpnode->pnode_link;
		}
	}
	else if(gnprefix_len==support)
	{
		while(ptid_node!=NULL)
		{
			if(gpitem_order_map[ptid_node->ntid]>gnfirstitem_order)
				return 0;
			ptid_list[nlist_len++] = ptid_node->ntid;
			if(ptid_node==pheader_node->ptid_tail)
				break;
			ptid_node = ptid_node->pnext;
		}		
		pfpnode = pheader_node->pconddb;
		while(pfpnode!=NULL )
		{
			ptid_node = pfpnode->ptid_head;
			while(1)
			{
				if(gpitem_order_map[ptid_node->ntid]>gnfirstitem_order)
					return 0;
				ptid_list[nlist_len++] = ptid_node->ntid;
				if(ptid_node==pfpnode->ptid_tail)
					break;
				ptid_node = ptid_node->pnext;
			}
			pfpnode = pfpnode->pnode_link;
		}
	}
	else 
		printf("Error with length and support\n");

	if(nlist_len!=support)
		printf("Error: inconsistent support\n");

	return nlist_len;

}


//sort items in descending frequency order
int comp_item_freq_des(const void *e1, const void *e2)
{
	ITEM_COUNTER *p1, *p2;
	p1 = (ITEM_COUNTER *) e1;
	p2 = (ITEM_COUNTER *) e2;

	if ((p1->nsupport>p2->nsupport) )
		return -1;
	else if (p1->nsupport<p2->nsupport)
		return 1;
	else 
		return 0;
}


//sort items in descending frequency order
int comp_hnode_freq_des(const void *e1, const void *e2)
{
	HEADER_NODE *p1, *p2;
	p1 = (HEADER_NODE *) e1;
	p2 = (HEADER_NODE *) e2;

	if ((p1->nsupport>p2->nsupport) )
		return -1;
	else if (p1->nsupport<p2->nsupport)
		return 1;
	else if(p1->nitem>p2->nitem)
		return -1;
	else if(p1->nitem<p2->nitem)
		return 1;
	else
		return 0;
}
