#include <stdlib.h>
#include <stdio.h>

#include "ScanDBMine.h"
#include "data.h"
#include "inline_routine.h"

CScanDBMine  goDBMiner;


// Enumerate all the frequent items in the database
int CScanDBMine::ScanDBCountFreqItems(int* &pitem_sup_map)
{
	Transaction* ptransaction;
	int *ptemp_map, ncapacity, nnewcapacity;
	int i;

	ncapacity = ITEM_SUP_MAP_SIZE;
	pitem_sup_map = NewIntArray(ncapacity, 0);

	gndb_size = 0;
	gnmax_trans_len = 0;
	gnmax_item_id = 0;

	ptransaction=gpdata->getNextTransaction();
	while(ptransaction)
	{
		if(ptransaction->length>0)
		{
			if(gnmax_trans_len<ptransaction->length)
				gnmax_trans_len = ptransaction->length;
			if(ptransaction->length>=gnmin_size)
			{
				for(i=0;i<ptransaction->length;i++)
				{
					if(gnmax_item_id<ptransaction->t[i])
						gnmax_item_id = ptransaction->t[i];

					if(ptransaction->t[i]>=ncapacity)
					{
						nnewcapacity = MAX(2*ncapacity, ptransaction->t[i]+1);
						ptemp_map = NewIntArray(nnewcapacity);
						memcpy(ptemp_map, pitem_sup_map, sizeof(int)*ncapacity);
						memset(&(ptemp_map[ncapacity]), 0, sizeof(int)*(nnewcapacity-ncapacity));
						DelIntArray(pitem_sup_map, ncapacity);
						pitem_sup_map = ptemp_map;
						ncapacity = nnewcapacity;
					}
					pitem_sup_map[ptransaction->t[i]]++;
				}
			}
		}
		gndb_size++;
		ptransaction = gpdata->getNextTransaction();
	}

	gnmax_trans_len++;
	gnmax_item_id++;

	return ncapacity;
}

FP_NODE* CScanDBMine::ScanDBBuildFPtree(HEADER_TABLE pheader_table, int *pitem_order_map)
{
	Transaction *ptransaction;
	int ntrans_len, i, order, ntid;
	FP_NODE *proot;
	TID_NODE *ptid_node;

	ntid = 0;
	proot = NULL;
	ptransaction = gpdata->getNextTransaction();
	while(ptransaction)
	{
		if(ptransaction->length>=gnmin_size)
		{
			ntrans_len = 0;
			for(i=0;i<ptransaction->length;i++)
			{
				order = pitem_order_map[ptransaction->t[i]];
				if(order>=0)
				{
					gptransaction[ntrans_len] = order;
					ntrans_len++;
				}
			}

			if(ntrans_len>=gnmin_size-gnprefix_len)
			{
				if(ntrans_len>1)
				{
					//qsort((void*)gptransaction, ntrans_len, sizeof(int), comp_int_asc);
					sort_trans(gptransaction, ntrans_len);
					ptid_node = NewTidNode(ntid);
					goFPtree.InsertTransaction(proot, pheader_table, gptransaction, ntrans_len, ntrans_len-1, 1, ptid_node, ptid_node);
				}
				else if(ntrans_len==1)
				{
					ptid_node = NewTidNode(ntid);
					if(pheader_table[gptransaction[0]].ptid_head==NULL)
						pheader_table[gptransaction[0]].ptid_head = ptid_node;
					else 
						pheader_table[gptransaction[0]].ptid_tail->pnext = ptid_node;
					pheader_table[gptransaction[0]].ptid_tail = ptid_node;
				}
			}
		}

		ntid++;
		ptransaction = gpdata->getNextTransaction();
	}

	return proot;
}


int CScanDBMine::ScanDBGetTidList(int nitem, int* ptid_list)
{
	Transaction* ptransaction;
	int i, nlist_len, ntid;

	nlist_len = 0;
	ntid = 0;
	ptransaction = gpdata->getNextTransaction();
	while(ptransaction)
	{
		if(ptransaction->length>0)
		{
			for(i=0;i<ptransaction->length;i++)
			{
				if(ptransaction->t[i]==nitem)
				{
					ptid_list[nlist_len++] = ntid;
					break;
				}	
			}
		}
		ntid++;
		ptransaction = gpdata->getNextTransaction();
	}
	return nlist_len;
}


// Enumerate all the frequent items in the database
int CScanDBMine::ScanGraphCountFreqVertices(int* &pitem_sup_map)
{
	Transaction* ptransaction;
	int *ptemp_map, ncapacity, nnewcapacity, ntid, i;

	ncapacity = ITEM_SUP_MAP_SIZE;
	pitem_sup_map = NewIntArray(ncapacity, 0);

	gnmax_trans_len = 0;

	ntid = 0;
	ptransaction=gpdata->getNextTransaction();
	while(ptransaction)
	{
		if(ptransaction->length>0)
		{
			if(gnmax_trans_len<ptransaction->length)
				gnmax_trans_len = ptransaction->length;
			if(ptransaction->length>=gnmin_size)
			{
				if(ntid>=ncapacity)
				{
					ptemp_map = NewIntArray(2*ncapacity);
					memcpy(ptemp_map, pitem_sup_map, sizeof(int)*ncapacity);
					memset(&(ptemp_map[ncapacity]), 0, sizeof(int)*ncapacity);
					DelIntArray(pitem_sup_map, ncapacity);
					pitem_sup_map = ptemp_map;
					ncapacity *= 2; 
				}
				pitem_sup_map[ntid] = ptransaction->length;
			}
			else
			{
				for(i=0;i<ptransaction->length;i++)
				{
					if(ptransaction->t[i]>=ncapacity)
					{
						nnewcapacity = MAX(2*ncapacity, ptransaction->t[i]+1);
						ptemp_map = NewIntArray(nnewcapacity);
						memcpy(ptemp_map, pitem_sup_map, sizeof(int)*ncapacity);
						memset(&(ptemp_map[ncapacity]), 0, sizeof(int)*(nnewcapacity-ncapacity));
						DelIntArray(pitem_sup_map, ncapacity);
						pitem_sup_map = ptemp_map;
						ncapacity = nnewcapacity;
					}
					pitem_sup_map[ptransaction->t[i]]--;
				}
			}
		}
		ntid++;
		ptransaction = gpdata->getNextTransaction();
	}

	gnmax_trans_len++;
	gndb_size = ntid;
	gnmax_item_id = ntid;

	return ncapacity;
}

int CScanDBMine::ScanGraphGetTidList(int nvertex, int* ptid_list)
{
	Transaction* ptransaction;
	int nlist_len, ntid;

	nlist_len = 0;
	ntid = 0;
	ptransaction = gpdata->getNextTransaction();
	while(ptransaction)
	{
		if(ntid==nvertex && ptransaction->length>0)
		{
			memcpy(ptid_list, ptransaction->t, sizeof(int)*ptransaction->length);
			nlist_len = ptransaction->length;
			break;
		}
		ntid++;
		ptransaction = gpdata->getNextTransaction();
	}
	return nlist_len;
}



void CScanDBMine::PrintSum(FILE *fp)
{
//	fprintf(fp, "%d %d %d\t", gndb_size, mnadj_list_capacity, mnadj_list_size);
	fprintf(fp, "%d %d %d\t", gntotal_freqitems, mnpruned_items, mninit_pruned_items);
}

void CScanDBMine::LoadDB(int* pitem_sup_map, int nsup_sum)
{
	Transaction *ptransaction;
	int i, ntrans_len, ntid, nbuf_pos;

	mngraph_size = nsup_sum+gndb_size;
	mpgraph = NewIntArray(mngraph_size);
	mppadj_lists = NewIntPointerArray(gndb_size);
	memset(mppadj_lists, 0, sizeof(int*)*gndb_size);

	ntid = 0;
	nbuf_pos = 0;
	ptransaction = gpdata->getNextTransaction();
	while(ptransaction)
	{
		if(ptransaction->length>0)
		{
			if(ptransaction->length+gnprefix_len>=gnmin_size)
			{
				mppadj_lists[ntid] = &(mpgraph[nbuf_pos]);
				ntrans_len = 1;
				for(i=0;i<ptransaction->length;i++)
				{
					if(pitem_sup_map[ptransaction->t[i]])
						mppadj_lists[ntid][ntrans_len++] = ptransaction->t[i];
				}
				ntrans_len--;
				if(ntrans_len>0 && ntrans_len+gnprefix_len>=gnmin_size)
				{
					mppadj_lists[ntid][0] = ntrans_len;
					nbuf_pos += ntrans_len+1;
				}
				else
					mppadj_lists[ntid] = NULL;
			}
		}
		ntid++;
		ptransaction = gpdata->getNextTransaction();
	}
}


void CScanDBMine::DestroyDB()
{
	DelIntPointerArray(mppadj_lists, gndb_size);
	DelIntArray(mpgraph, mngraph_size);
}