#pragma once 

#include <string.h>
#include "Global.h"

/*
// for detecting memory leak
#include <afx.h>
#define new DEBUG_NEW
#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif
//*/


inline void IncMemSize(int size)
{
	gdused_mem_size += size;
	if(gdused_mem_size>gdmax_used_mem_size)
		gdmax_used_mem_size = gdused_mem_size;
}
inline void DecMemSize(int size)
{
	gdused_mem_size -= size;
}



//=============================================================================
inline int* NewIntArray(int length)
{
	IncMemSize(length*sizeof(int));
	return new int[length];
}
inline void DelIntArray(int* parray, int length)
{
	DecMemSize(length*sizeof(int));
	delete []parray;
}
inline int** NewIntPointerArray(int length)
{
	IncMemSize(length*sizeof(int*));
	return new int*[length];
}
inline void DelIntPointerArray(int** ppitemsets, int length)
{
	DecMemSize(length*sizeof(int*));
	delete []ppitemsets;
}
inline int* NewIntArray(int length, int value)
{
	int *parray;

	parray = NewIntArray(length);
	memset(parray, value, sizeof(int)*length);
	return parray;
}
inline char* NewCharArray(int length)
{
	IncMemSize(length*sizeof(char));
	return new char[length];
}
inline void DelCharArray(char* parray, int length)
{
	DecMemSize(length*sizeof(char));
	delete []parray;
}
inline unsigned short* NewUShortArray(int length)
{
	IncMemSize(length*sizeof(unsigned short));
	return new unsigned short[length];
}
inline void DelUShortArray(unsigned short* parray, int length)
{
	DecMemSize(length*sizeof(unsigned short));
	delete []parray;
}

//=============================================================================


//-----------------------------------------------------------------------------
#include "FPtree.h"

#define FPNODE_PAGE_SIZE (1<<10)

struct FPNODE_PAGE
{
	FP_NODE pfpnodes[FPNODE_PAGE_SIZE];
	FPNODE_PAGE *pnext;
};
struct FPNODE_BUF
{
	FPNODE_PAGE *phead;
	FPNODE_PAGE *pcur_page; 
	int ncur_pos;
	int ntotal_pages;
};
extern FPNODE_BUF gofpnode_buf;

inline FPNODE_PAGE* NewFPNodePage()
{
	FPNODE_PAGE *pfppage;

	pfppage = new FPNODE_PAGE;
	IncMemSize(sizeof(FPNODE_PAGE));
	pfppage->pnext = NULL;

	return pfppage;
}
inline void DelFFNodePage(FPNODE_PAGE* pfppage)
{
	DecMemSize(sizeof(FPNODE_PAGE));
	delete pfppage;
}

inline FP_NODE* NewOneFPNode()
{
	FP_NODE *pfpnode;
	FPNODE_PAGE *pnewfppage;

	if(gofpnode_buf.ncur_pos==FPNODE_PAGE_SIZE)
	{
		if(gofpnode_buf.pcur_page->pnext==NULL)
		{
			pnewfppage = NewFPNodePage();
			gofpnode_buf.pcur_page->pnext = pnewfppage;
			gofpnode_buf.pcur_page = pnewfppage;
			gofpnode_buf.ntotal_pages++;
		}
		else 
			gofpnode_buf.pcur_page = gofpnode_buf.pcur_page->pnext;
		gofpnode_buf.ncur_pos = 0;
	}
	
	pfpnode = &(gofpnode_buf.pcur_page->pfpnodes[gofpnode_buf.ncur_pos]);
	gofpnode_buf.ncur_pos++;
	pfpnode->ptid_head = NULL;
	pfpnode->ptid_tail = NULL;
	pfpnode->ptail_itemset = NULL;
	
	return pfpnode;
}

extern TID_NODE *gptid_nodes;

inline TID_NODE* NewTidNodeArray(int length)
{
	IncMemSize(sizeof(TID_NODE)*length);
	return new TID_NODE[length];
}

inline TID_NODE* NewTidNode(int ntid)
{
	gptid_nodes[ntid].ntid = ntid;
	gptid_nodes[ntid].pnext = NULL;
	return &(gptid_nodes[ntid]);

}
inline void DelTidNodeArray(TID_NODE *ptid_nodes, int length)
{
	DecMemSize(sizeof(TID_NODE)*length);
	delete []ptid_nodes;
}

//--------------------------------------------------------------------------



//=============================================================================
extern HEADER_TABLE gpdfs_header_array;
extern int gndfs_header_size;
extern int gndfs_header_pos;

inline HEADER_TABLE NewHeaderTable(int num_of_freqitems)
{
	HEADER_TABLE pheader_table;

	if(gndfs_header_pos+num_of_freqitems<=gndfs_header_size)
	{
		pheader_table = &(gpdfs_header_array[gndfs_header_pos]);
		gndfs_header_pos += num_of_freqitems;
	}
	else 
	{
		IncMemSize(num_of_freqitems*sizeof(HEADER_NODE));
		pheader_table = new HEADER_NODE[num_of_freqitems];
	}

	return pheader_table;
}

inline void DelHeaderTable(HEADER_TABLE pheader_table, int num_of_freqitems)
{
	if((unsigned int)pheader_table>=(unsigned int)gpdfs_header_array && (unsigned int)pheader_table<(unsigned int)gpdfs_header_array+sizeof(HEADER_NODE)*gndfs_header_size)
	{
		gndfs_header_pos -= num_of_freqitems;
	}
	else 
	{
		DecMemSize(num_of_freqitems*sizeof(HEADER_NODE));
		delete []pheader_table;
	}

}

inline ITEM_COUNTER* NewItemCounter(int length)
{
	IncMemSize(sizeof(ITEM_COUNTER)*length);
	return new ITEM_COUNTER[length];
}
inline void DelItemCounter(ITEM_COUNTER* pcounters, int length)
{
	DecMemSize(sizeof(ITEM_COUNTER)*length);
	delete []pcounters;
}


//=============================================================================


//------------------------------------------------------------------------------
#define TAIL_BUF_PAGE_SIZE  1024
#define TAIL_BUF_PAGES   1024
extern int** gpptailitem_buf;
extern int gntail_buf_pages;
extern int gntail_buf_size;
extern int gnpage_no;
extern int gnpage_pos;

inline void InitTailBuf()
{
	gntail_buf_pages = TAIL_BUF_PAGES;
	gpptailitem_buf = NewIntPointerArray(gntail_buf_pages);
	memset(gpptailitem_buf, 0, sizeof(int*)*gntail_buf_pages);
	gpptailitem_buf[0] = NewIntArray(TAIL_BUF_PAGE_SIZE);

	gnpage_no = 0;
	gnpage_pos = 0;
}

inline void DelTailBuf()
{
	int i;

	gntail_buf_size = 0;
	for(i=0;i<gntail_buf_pages;i++)
	{
		if(gpptailitem_buf[i]!=NULL)
		{
			DelIntArray(gpptailitem_buf[i], TAIL_BUF_PAGE_SIZE);
			gntail_buf_size += TAIL_BUF_PAGE_SIZE*sizeof(int);
		}
	}
	gntail_buf_size += sizeof(int*)*gntail_buf_pages;
	DelIntPointerArray(gpptailitem_buf, gntail_buf_pages);
}

inline int* InsertTailItemset(int length)
{
	int *pitemset; 

	if(gnpage_pos+length+1>TAIL_BUF_PAGE_SIZE)
	{
		if(gnpage_no+1==gntail_buf_pages)
		{
			int ** pptemp;

			pptemp = NewIntPointerArray(gntail_buf_pages*2);
			memcpy(pptemp, gpptailitem_buf, sizeof(int*)*gntail_buf_pages);
			memset(&(pptemp[gntail_buf_pages]), 0, sizeof(int*)*gntail_buf_pages);
			DelIntPointerArray(gpptailitem_buf, gntail_buf_pages);
			gpptailitem_buf = pptemp;

			gnpage_no++;
			gntail_buf_pages *= 2;
		}
		else
			gnpage_no++;

		if(gpptailitem_buf[gnpage_no]==NULL)
			gpptailitem_buf[gnpage_no] = NewIntArray(TAIL_BUF_PAGE_SIZE);
		gnpage_pos = 0;
	}

	pitemset = &(gpptailitem_buf[gnpage_no][gnpage_pos]);
	gnpage_pos += length+1;

	return pitemset;
}


//------------------------------------------------------------------------------


//=============================================================================
inline void Init()
{
	gpsort_trans_buf = NewCharArray(gntotal_freqitems);
	memset(gpsort_trans_buf, 0, sizeof(char)*gntotal_freqitems);


	//for dfs mining
	gndfs_header_size = gntotal_freqitems*3;
	gpdfs_header_array = new HEADER_NODE[gndfs_header_size];
	IncMemSize(sizeof(HEADER_NODE)*gndfs_header_size);
	gndfs_header_pos = 0;


    // for allocating space for fptree
	gofpnode_buf.phead = NewFPNodePage();
	gofpnode_buf.pcur_page = gofpnode_buf.phead;
	gofpnode_buf.ncur_pos = 0;
	gofpnode_buf.ntotal_pages = 1;	
}

inline void Reset(FPNODE_PAGE *pcur_page, int ncur_pos)
{
	gofpnode_buf.pcur_page = pcur_page;
	gofpnode_buf.ncur_pos = ncur_pos;
}

inline void Destroy()
{
	DelCharArray(gpsort_trans_buf, gntotal_freqitems);

	//for DFS mining
	DecMemSize(sizeof(HEADER_NODE)*gndfs_header_size);
	delete []gpdfs_header_array;

	//releasing fpnode buffer
	FPNODE_PAGE *pfppage, *pnext_fppage;
	pfppage = gofpnode_buf.phead;
	while(pfppage!=NULL)
	{
		pnext_fppage = pfppage->pnext;
		DelFFNodePage(pfppage);
		pfppage = pnext_fppage;	
		gofpnode_buf.ntotal_pages--;
	}
	if(gofpnode_buf.ntotal_pages!=0)
		printf("Error with number of FPNode pages\n");
}
//=============================================================================




inline void sort_trans(int *ptransaction, int ntrans_len)
{
	int i, j;

	if(gnum_of_newfreqitems/ntrans_len>64)
		qsort(ptransaction, ntrans_len, sizeof(int), comp_int_asc);
	else
	{
		for(i=0;i<ntrans_len;i++)
			gpsort_trans_buf[ptransaction[i]] = 1;

		j = 0;
		for(i=0;i<gnum_of_newfreqitems;i++)
		{
			if(gpsort_trans_buf[i])
			{
				ptransaction[j] = i;
				j++;
				gpsort_trans_buf[i] = 0;
			}
		}
	}

}


