#include <stdlib.h>
//#include <crtdbg.h>

#include <string.h>
#include <time.h>
#include <sys/timeb.h>

#include "ScanDBMine.h"
#include "FPtree.h"
#include "fsout.h"
#include "data.h"
#include "inline_routine.h"

void MineBP_Graph()
{
	FP_NODE *proot;
	HEADER_TABLE pheader_table;
	int *pitem_sup_map, ncapacity, i, nfreqsup_sum, nfreq_item, nmax_sup, nmax_sup_item, nmin_sup, nend_pos;
	struct timeb start, end;

	ftime(&start);

	if(gbresult_name_given)
		gpfout = new FSout(gszresult_filename);

	gpdata = new Data(gszdata_filename);

	gnmax_pattern_len = 0;
	gntotal_call = 0;
	gncountsup_calls = 0;
	gncounttail_calls = 0;
	gdtail_count_time = 0;
	gdused_mem_size = 0;
	gdmax_used_mem_size = 0;
	gntree_max_size = 0;
	gntotal_closed = 0;
	gntotal_bipartite = 0;

	if(gnmin_sup<gnmin_size)
	{
		i = gnmin_sup;
		gnmin_sup = gnmin_size;
		gnmin_size = i;
	}

	//count frequent items in original database
	pitem_sup_map = NULL;
	ncapacity = goDBMiner.ScanGraphCountFreqVertices(pitem_sup_map);

	gpprefix_itemset = NewIntArray(gnmax_trans_len);
	gnprefix_len = 0;

	//enumerate frequent itemsets
	gntotal_freqitems  = 0;
	nmax_sup = 0;
	nmax_sup_item = -1;
	nmin_sup = gndb_size;
	nfreqsup_sum = 0;
	for(i=0;i<gnmax_item_id;i++)
	{
		if(pitem_sup_map[i]==gndb_size)
		{
			gpprefix_itemset[gnprefix_len++] = i;
			pitem_sup_map[i] = 0;
		}
		else if(pitem_sup_map[i]>=gnmin_sup)
		{
			gntotal_freqitems++;
			nfreq_item = i;
			if(nmax_sup < pitem_sup_map[i])
			{
				nmax_sup = pitem_sup_map[i];
				nmax_sup_item = i;
			}
			if(nmin_sup>pitem_sup_map[i])
				nmin_sup = pitem_sup_map[i];
			nfreqsup_sum += pitem_sup_map[i];
		}
		else
			pitem_sup_map[i] = 0;
	}
	if(gnprefix_len>=gnmin_size)
	{
		gptid_list = NULL;
		OutputOneClosedPat(gndb_size, gptid_list);
	}

	if(gnprefix_len+gntotal_freqitems<gnmin_size)
	{
		delete gpdata;
		delete gpfout;
		DelIntArray(gpprefix_itemset, gnmax_trans_len);
		DelIntArray(pitem_sup_map, ncapacity);
		return;
	}

	gnfirstitem_pos = gnprefix_len;
	gptransaction = NewIntArray(gnmax_trans_len);

	gptid_list = NewIntArray(nmax_sup);

	if(gntotal_freqitems<=1)
	{
		if(gntotal_freqitems==1 && gnprefix_len+1>=gnmin_size)
		{
			goDBMiner.ScanGraphGetTidList(nfreq_item, gptid_list);
			gpprefix_itemset[gnprefix_len++] = nfreq_item;
			OutputOneClosedPat(pitem_sup_map[nfreq_item], gptid_list);
			gnprefix_len--;
		}
		gnmax_pattern_len = gnprefix_len+1;
		DelIntArray(pitem_sup_map, ncapacity);
	}
	else if(gntotal_freqitems>1 && gntotal_freqitems+gnprefix_len>=gnmin_size)
	{
		gplocal_item_sup_map = NewIntArray(gntotal_freqitems);
		gplocal_item_sumsup_map = NewIntArray(gntotal_freqitems);
		InitTailBuf();

		//create header table
		pheader_table = NewHeaderTable(gntotal_freqitems);
		gntotal_freqitems = 0;
		for(i=0;i<gnmax_item_id;i++)
		{
			if(pitem_sup_map[i]>=gnmin_sup && pitem_sup_map[i]<gndb_size)
			{
				pheader_table[gntotal_freqitems].nitem = i;
				pheader_table[gntotal_freqitems].nsupport = pitem_sup_map[i];
				pheader_table[gntotal_freqitems].pconddb = NULL;
				pheader_table[gntotal_freqitems].ptid_head = NULL;
				pheader_table[gntotal_freqitems].ptid_tail = NULL;
				pheader_table[gntotal_freqitems].bgen_fullitems = false;
				pheader_table[gntotal_freqitems].bisclosed = true;
				gntotal_freqitems++;
			}
		}
		if(nmax_sup>nmin_sup)
			qsort(pheader_table, gntotal_freqitems, sizeof(HEADER_NODE), comp_hnode_freq_des);
		gpheader_table = pheader_table;
		DelIntArray(pitem_sup_map, ncapacity);


		//build tree
		gpitem_order_map = NewIntArray(gnmax_item_id, -1);
		for(i=0;i<gntotal_freqitems;i++)
			gpitem_order_map[pheader_table[i].nitem] =i;

		Init();
		gptid_nodes = NewTidNodeArray(gndb_size);

		gnum_of_newfreqitems = gntotal_freqitems;

		proot = goDBMiner.ScanDBBuildFPtree(pheader_table, gpitem_order_map);
		gntree_init_size = sizeof(FPNODE_BUF)+gofpnode_buf.ntotal_pages*sizeof(FPNODE_PAGE);
		

		//depth first mining
		gpdfs_item_suporder_map = NewIntArray(gntotal_freqitems);
		gpdfs_item_counter = NewItemCounter(gntotal_freqitems);
		gpdfs_cofreqs = NewIntArray(gntotal_freqitems*(gntotal_freqitems+1)/2);
		gpdfs_pruned_items = NewIntArray(gntotal_freqitems);

		nend_pos = gnmin_size-gnprefix_len-1;
		if(nend_pos<0)
			nend_pos = 0;
		goFPtree.GraphDepthFGGrowth(proot, pheader_table, gntotal_freqitems, 0, nend_pos);

		gntree_max_size = sizeof(FPNODE_BUF)+gofpnode_buf.ntotal_pages*sizeof(FPNODE_PAGE);
		DelIntArray(gpdfs_item_suporder_map, gntotal_freqitems);
		DelItemCounter(gpdfs_item_counter, gntotal_freqitems);
		DelIntArray(gpdfs_cofreqs, gntotal_freqitems*(gntotal_freqitems+1)/2);
		DelIntArray(gpdfs_pruned_items, gntotal_freqitems);
		
		DelTidNodeArray(gptid_nodes, gndb_size);
		DelIntArray(gpitem_order_map, gnmax_item_id);
		DelHeaderTable(pheader_table, gntotal_freqitems);
		Destroy();

		DelTailBuf();
		DelIntArray(gplocal_item_sup_map, gntotal_freqitems);
		DelIntArray(gplocal_item_sumsup_map, gntotal_freqitems);
	}
	else
	{
		DelIntArray(pitem_sup_map, ncapacity);
	}
	DelIntArray(gptid_list, nmax_sup);

	DelIntArray(gpprefix_itemset, gnmax_trans_len);
	DelIntArray(gptransaction, gnmax_trans_len);

	delete gpdata;
	if(gbresult_name_given)
		delete gpfout;
	ftime(&end);
	gdtotal_running_time = end.time-start.time+(double)(end.millitm-start.millitm)/1000;

}


void FPtree::GraphDepthFGGrowth(FP_NODE *proot, HEADER_TABLE pheader_table, int num_of_freqitems, int num_of_tail_items, int nend_pos)
{
	HEADER_TABLE pnewheader_table;
	FP_NODE *pnewroot, *pfpnode;
	FPNODE_PAGE *pstart_page;
	int k, i, nstart_pos, ncount, nlist_len, nnew_end_pos, ntailbuf_page_no, ntailbuf_page_pos;
	int num_of_tail_comm, num_of_comm_items, num_of_newfreqitems, num_of_newtail_items, num_of_newtail_comm;

	gntotal_call++;

	for(k=num_of_freqitems-1;k>=nend_pos; k--)
	{
		if(pheader_table[k].bisclosed)
		{
			if(gnprefix_len==gnfirstitem_pos)
				gnfirstitem_order = k;
			gpprefix_itemset[gnprefix_len] = pheader_table[k].nitem;
			gnprefix_len++;

			if(pheader_table[k].pconddb==NULL)
			{
				if(gnprefix_len>=gnmin_size && gnprefix_len<=pheader_table[k].nsupport)
				{
					nlist_len = GetTidList(pheader_table[k].nsupport, &(pheader_table[k]), gptid_list);
					OutputOneClosedPat(nlist_len, gptid_list);
				}
			}
			else if(gnprefix_len<=pheader_table[k].nsupport)
			{
				if(pheader_table[k].pconddb->pnode_link==NULL)
				{
					ncount = pheader_table[k].pconddb->frequency;
					if(ncount<pheader_table[k].nsupport)
					{
						if(gnprefix_len>=gnmin_size)
						{
							nlist_len = GetTidList(pheader_table[k].nsupport, &(pheader_table[k]), gptid_list);
							OutputOneClosedPat(nlist_len, gptid_list);
						}
					}
					if(ncount>=gnmin_sup)
					{
						num_of_comm_items = 0;
						pfpnode = pheader_table[k].pconddb->pparent;
						while(pfpnode!=NULL)
						{
							gpprefix_itemset[gnprefix_len++] = pheader_table[pfpnode->nitem_order].nitem;
							if(pheader_table[pfpnode->nitem_order].nsupport==ncount)
								pheader_table[pfpnode->nitem_order].bisclosed = false;
							num_of_comm_items++;
							pfpnode = pfpnode->pparent;
						}
						if(gnprefix_len>=gnmin_size && gnprefix_len<=ncount)
						{
							num_of_tail_comm = CountTailFullItems(pheader_table, k, ncount, num_of_freqitems, num_of_freqitems+num_of_tail_items);
							if(num_of_tail_comm==0)
							{
								if(num_of_comm_items>0 || ncount==pheader_table[k].nsupport)
								{
									nlist_len = GetTidList(ncount, pheader_table[k].pconddb, gptid_list);
									OutputOneClosedPat(nlist_len, gptid_list);
								}
							}
						}
						gnprefix_len -= num_of_comm_items;
					}
				}
				else
				{
					//count frequent items from FP-tree
					memset(gpdfs_item_suporder_map, 0, sizeof(int)*k);
					memset(&(gpdfs_item_suporder_map[k]), -1, sizeof(int)*(num_of_freqitems+num_of_tail_items-k));
					
					CountFreqItems(pheader_table, k, gpdfs_item_suporder_map);

					num_of_newfreqitems = 0;
					num_of_newtail_items = 0;
					num_of_comm_items = 0;
					for(i=0;i<k;i++)
					{
						if(gpdfs_item_suporder_map[i]==pheader_table[k].nsupport)
						{
							gpprefix_itemset[gnprefix_len++] = pheader_table[i].nitem;
							if(pheader_table[i].nsupport==pheader_table[k].nsupport)
								pheader_table[i].bisclosed = false;
							num_of_comm_items++;
						}
						else if(gpdfs_item_suporder_map[i]>=gnmin_sup && gpdfs_item_suporder_map[i]>gnprefix_len)
						{
							gpdfs_item_counter[num_of_newfreqitems].nsupport = gpdfs_item_suporder_map[i];
							gpdfs_item_counter[num_of_newfreqitems].nitem = pheader_table[i].nitem;
							gpdfs_item_counter[num_of_newfreqitems].norder = i;
							num_of_newfreqitems++;
						}
						gpdfs_item_suporder_map[i] = -1;
					}

					if(num_of_newfreqitems+gnprefix_len>=gnmin_size && gnprefix_len<=pheader_table[k].nsupport)
					{
						if(num_of_tail_items>0 || k<num_of_freqitems-1)
							num_of_tail_comm = CountTailFullItems(pheader_table, k, pheader_table[k].nsupport, num_of_freqitems, num_of_freqitems+num_of_tail_items);
						else
						{
							pheader_table[k].bgen_fullitems = true;
							num_of_tail_comm = 0;
						}

						if(num_of_tail_comm==0)
						{
							if(gnprefix_len>=gnmin_size)
							{
								nlist_len = GetTidList(pheader_table[k].nsupport, &(pheader_table[k]), gptid_list);
								OutputOneClosedPat(nlist_len, gptid_list);
							}

							if(gnprefix_len<pheader_table[k].nsupport)
							{
								if(num_of_newfreqitems>1)
								{
									qsort(gpdfs_item_counter, num_of_newfreqitems, sizeof(ITEM_COUNTER), comp_item_freq_des);
									for(i=0;i<num_of_newfreqitems;i++)
										gpdfs_item_suporder_map[gpdfs_item_counter[i].norder] = i;
								}

								for(i=k+1;i<num_of_freqitems+num_of_tail_items;i++)
								{
									if(gplocal_item_sumsup_map[i]>=gnmin_sup && gplocal_item_sumsup_map[i]>gnprefix_len && pheader_table[i].bisclosed)
									{
										gpdfs_item_suporder_map[i] = num_of_newfreqitems+num_of_newtail_items;
										gpdfs_item_counter[num_of_newfreqitems+num_of_newtail_items].nitem = pheader_table[i].nitem;
										gpdfs_item_counter[num_of_newfreqitems+num_of_newtail_items].nsupport = 0;
										num_of_newtail_items++;
									}
								}

								if(num_of_newfreqitems==1)
								{
									gpprefix_itemset[gnprefix_len++] = gpdfs_item_counter[0].nitem;
									if(gnprefix_len<=gpdfs_item_counter[0].nsupport)
									{
										num_of_newtail_comm = GetTailFullItems(pheader_table[k].pconddb, gpdfs_item_counter[0].nsupport, gpdfs_item_counter[0].norder, num_of_freqitems+num_of_tail_items);
										if(num_of_newtail_comm==0)
										{
											nlist_len = GetTidList(gpdfs_item_counter[0].norder, gpdfs_item_counter[0].nsupport, pheader_table[k].pconddb, gptid_list);
											OutputOneClosedPat(nlist_len, gptid_list);
										}
									}
									gnprefix_len--;
								}
								else if(num_of_newfreqitems>1)
								{
									pnewheader_table = NewHeaderTable(num_of_newfreqitems+num_of_newtail_items);
									for(i=0;i<num_of_newfreqitems+num_of_newtail_items;i++)
									{
										pnewheader_table[i].nitem = gpdfs_item_counter[i].nitem;
										pnewheader_table[i].nsupport = gpdfs_item_counter[i].nsupport;
										pnewheader_table[i].pconddb = NULL;
										pnewheader_table[i].ptid_head = NULL;
										pnewheader_table[i].ptid_tail = NULL;
										pnewheader_table[i].bgen_fullitems = false;
										pnewheader_table[i].bisclosed = true;
									}

									gnum_of_newfreqitems = num_of_newfreqitems+num_of_newtail_items;
									pstart_page = gofpnode_buf.pcur_page;
									nstart_pos = gofpnode_buf.ncur_pos;
									ntailbuf_page_no = gnpage_no;
									ntailbuf_page_pos = gnpage_pos;
									
									pnewroot = BuildNewTreeWTTail(pheader_table[k].pconddb, pnewheader_table, gpdfs_item_suporder_map); 

									nnew_end_pos = gnmin_size-gnprefix_len-1;
									if(nnew_end_pos<0)
										nnew_end_pos = 0;
									GraphDepthFGGrowth(pnewroot, pnewheader_table, num_of_newfreqitems, num_of_newtail_items, nnew_end_pos);

									gnpage_no = ntailbuf_page_no;
									gnpage_pos = ntailbuf_page_pos;

									Reset(pstart_page, nstart_pos);
									DelHeaderTable(pnewheader_table, num_of_newfreqitems+num_of_newtail_items);
								}
							}
						}
					}
					gnprefix_len -= num_of_comm_items;
				}
			}
			gnprefix_len--;
		}
		if(k>0 && pheader_table[k].pconddb!=NULL)
			PushRightTidList(pheader_table[k].pconddb);
	}
}

